//Part of XDB Core source code.
//Version 1.2
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html


//object representing column
unit ColLevel;

interface

uses UserStructs, AbstractLayer, UserLevel, Stream64, DataObj;

type
   TXDBCol = class(TAXDBColumn)
             private
                FID:Int64;
                procedure GenerateID; //generates new valid ID into the FID
                                      //if it is not possible calls Abort
             protected
                FKind:Int64;
                FOwner:TAXDBTable;
                FOnDestroy:TXDBColumnDestroy;
                FOnRename:TXDBColumnRename;
                FMinVal,
                FMaxVal,
                FDefVal:TXDBDataObj;
                procedure SetKind(NewKind:Int64); override;
                function GetKind:Int64; override;
                procedure ConvertData(From:Int64); //converts column's data
                        //from value From to FKind
               function GetMinVal:TXDBDataObject; override;
               function GetMaxVal:TXDBDataObject; override;
               function GetDefVal:TXDBDataObject; override;
               function GetOwner:TAXDBTable; override;
               function GetInMask:WideString; override;
               function GetOutMask:WideString; override;
             public
               constructor Create(AOwner:TAXDBTable); override;
               destructor Destroy; override;

               procedure Assign(NewDesign:TColumnDesignTX); override;
               procedure Assign(ACol:TXDBColumn); override;

               function VariableDataLength:boolean; override; 
                        //returns true when Kind doesn't sign data with
                        //hard length

               function DefinitionSize:Int64; 
                         //returns how much bytes are needed to store definition
                         //into db file
               procedure ReadDefFromStream(Stream:TStream64);
                        //loads definition from stream from position At
               procedure FlushDefIntoStream(Stream:TStream64);
                        //stores definition from stream from position At

               property ID:Int64 read FID;
             end;

implementation

uses SysUtils, mask, xdbRetCodes, math, routines;

constructor TXDBCol.Create(AOwner:TAXDBTable);
begin
  FOwner:=AOwner;
  FOnDestroy:=AOwner.ColumnDestroyHandler;
  FOnRename:=AOwner.ColumnRenameHandler;
  GenerateID;

  FMinVal:=TXDBDataObj.Create(Self);
  FMaxVal:=TXDBDataObj.Create(Self);
  FDefVal:=TXDBDataObj.Create(Self);
end;

destructor TXDBCol.Destroy;
begin
  if Assigned(FOnDestroy) then FOnDestroy(Self);

  FMinVal.Free;
  FMaxVal.Free;
  FDefVal.Free;

  inherited;
end;

procedure TXDBCol.SetKind(NewKind:Int64);
var oldKind:Int64;
begin
  oldKind:=FKind;
  if (NewKind<0) or (NewKind>=xdbdtDTDefined) then NewKind:=xdbdtRawData;
  FKind:=NewKind;
  
  ConvertData(oldKind);
end;

procedure TXDBCol.Assign(NewDesign:TColumnDesignTX);
var oldKind:Int64;
    OldInMask:WideString;
begin
  FOwner.Owner.Owner.SetLastError(xdbrcOK);

  //have a look if can rename ourself to NewDesign.Name
  if Assigned(FOnRename) and not(FOnRename(Self, NewDesign.Name)) then
     begin      //no
       FOwner.Owner.Owner.SetLastError(xdbrcColAlreadyExists);
       Abort;
       exit;
     end;

  //try to convert values from text representation
  //and backup important data
  oldKind:=FKind;
  oldInMask:=FInMask;
  Kind:=NewDesign.Kind;
  FInMask:=NewDesign.InputMask;
  try
    FMinVal.AsWideString:=NewDesign.MinValue;
    FMaxVal.AsWideString:=NewDesign.MaxValue;
    FDefVal.AsWideString:=NewDesign.InitialValue;
  except
    //data couldn't be converted => restore backuped data
    FKind:=oldKind;
    FInMask:=OldInMask;
    FOwner.Owner.Owner.SetLastError(xdbrcInvalidDColDsgnData);
    Abort;
    exit;
  end;

  //conversion was succesfull => set remain attributes
  Comment:=NewDesign.Comment;
  Name:=NewDesign.Name;
  OutputMask:=NewDesign.OutputMask;
  MinLength:=max(0, NewDesign.MinLength);
  MaxLength:=max(0, NewDesign.MaxLength);

  //at the end convert data of column
  ConvertData(oldKind);
end;

procedure TXDBCol.Assign(ACol:TXDBColumn);
begin
  FOwner.Owner.Owner.SetLastError(xdbrcUnexpected);

  FKind:=ACol.Kind;
  FComment:=ACol.Comment;
  FInMask:=ACol.InputMask;
  FOutMask:=ACol.OutputMask;
  FMinLength:=ACol.MinLength;
  FMaxLength:=ACol.MaxLength;

  FMinVal.Assign(ACol.MinValue);
  FMaxVal.Assign(ACol.MaxValue);
  FDefVal.Assign(ACol.InitialValue);

  FOwner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBCol.ConvertData(From:Int64);
begin
  if From = FKind then exit;
end;

procedure TXDBCol.GenerateID;
var i:integer;
    NextInc:boolean;
    pom:TXDBCol;
    firstID:Int64;
begin
  FOwner.Owner.Owner.SetLastError(xdbrcOK);
  PDateTime(@FID)^:=Now; //we should unique ID immediately
  firstID:=ID;          //remember ID we tried as first one
  repeat
    NextInc:=false;
    for i:=0 to FOwner.ColumnsCount-1 do
       begin
         pom:=(FOwner.Columns[i] as TXDBCol);
         if (pom<>nil) and (pom.ID = FID) and (Self <> pom) then
            begin
              NextInc:=true;
              break;
            end;
       end;

    if not NextInc then exit;//generated number is unique

    inc(FID);
  until FID = firstID-1; //try all
                //announce faialure
  FOwner.Owner.Owner.SetLastError(xdbrcNoMoreColsPossible);
  Abort;       //all ID occupied => exception // at IA-32 compatible can not be
end;


function TXDBCol.DefinitionSize:Int64;
begin
  result:=64+FMinVal.FileDataSize+FMaxVal.FileDataSize+FDefVal.FileDataSize
          +(Length(FName)+Length(FComment)+Length(FInMask)+Length(FOutMask))shl 1;
end;


procedure TXDBCol.ReadDefFromStream(Stream:TStream64);
begin
  //read kind of data
  Stream.Read(FKind, 8);
  //id of columnd
  Stream.Read(FID, 8);
  //column's name
  FName:=ReadWideString(Stream);
  //comment
  FComment:=ReadWideString(Stream);
  //input mask
  FInMask:=ReadWideString(Stream);
  //output mask
  FOutMask:=ReadWideString(Stream);
  //values
  FDefVal.LoadFromStream(Stream);
  FMinVal.LoadFromStream(Stream);
  FMaxVal.LoadFromStream(Stream);
  //min length
  Stream.Read(FMinLength,  8);
  //max length
  Stream.Read(FMaxLength, 8);
end;

procedure TXDBCol.FlushDefIntoStream(Stream:TStream64);
begin
  //write data kind
  Stream.Write(FKind, 8);
  //id of column
  Stream.Write(FID, 8);
  //column's name
  WriteWideString(Stream, FName);
  //comment
  WriteWideString(Stream, FComment);
  //input mask
  WriteWideString(Stream, FInMask);
  //output mask
  WriteWideString(Stream, FOutMask);
  //values
  FDefVal.StoreToStream(Stream);
  FMinVal.StoreToStream(Stream);
  FMaxVal.StoreToStream(Stream);
  //mini length
  Stream.Write(FMinLength,  8);
  //max length
  Stream.Write(FMaxLength, 8);
end;

function TXDBCol.VariableDataLength:boolean;
begin
  result:=not (FKind in [xdbdtInteger, xdbdtReal, xdbdtCurrency, xdbdtDateTime, xdbdtBoolean]);
end;

function TXDBCol.GetMinVal:TXDBDataObject;
begin
  result:=FMinVal;
end;

function TXDBCol.GetMaxVal:TXDBDataObject;
begin
  result:=FMaxVal;
end;

function TXDBCol.GetDefVal:TXDBDataObject;
begin
  result:=FDefVal;
end;

function TXDBCol.GetOwner:TAXDBTable;
begin
  result:=FOwner;
end;

function TXDBCol.GetKind:Int64;
begin
  result:=FKind;
end;

function TXDBCol.GetInMask:WideString;
begin
  result:=FInMask;
end;

function TXDBCol.GetOutMask:WideString;
begin
  result:=FOutMask;
end;


end.
